@@ -9,14 +9,19 @@
 
 #include "cryptoconfigmodule.h"
 #include "cryptoconfigmodule_p.h"
+#include "cryptoconfigentryreaderport_p.h"
 #include "directoryserviceswidget.h"
 #include "kdhorizontalline.h"
 #include "filenamerequester.h"
 
+#include "kleo/keyserverconfig.h"
+#include "utils/gnupg.h"
+
 #include <qgpgme/cryptoconfig.h>
 
 #include <KLineEdit>
 #include <KLocalizedString>
+#include <KMessageBox>
 #include "kleo_ui_debug.h"
 #include <utils/formatting.h>
 #include <QIcon>
@@ -39,9 +44,14 @@
 #include <QComboBox>
 #include <QGroupBox>
 
+#include <gpgme.h>
+
 #include <memory>
 #include <limits>
 #include <array>
+#include <set>
+
+#include <KLazyLocalizedString>
 
 using namespace Kleo;
 
@@ -73,11 +83,13 @@ static unsigned int num_components_with_options(const QGpgME::CryptoConfig *conf
     }
     const QStringList components = config->componentList();
     unsigned int result = 0;
-    for (QStringList::const_iterator it = components.begin(); it != components.end(); ++it)
-        if (const QGpgME::CryptoConfigComponent *const comp = config->component(*it))
+    for (QStringList::const_iterator it = components.begin(); it != components.end(); ++it) {
+        if (const QGpgME::CryptoConfigComponent *const comp = config->component(*it)) {
             if (!comp->groupList().empty()) {
                 ++result;
             }
+        }
+    }
     return result;
 }
 
@@ -212,7 +224,8 @@ QStringList sortConfigEntries(const Iterator orderBegin, const Iterator orderEnd
     // components sorting algorithm:
     // 1. components with predefined order (provided via orderBegin / orderEnd)
     // 2. other components sorted alphabetically
-    QStringList result, others;
+    QStringList result;
+    QStringList others;
     for (auto it = orderBegin; it != orderEnd; ++it) {
         if (entries.contains(*it)) {
             result.append(*it);
@@ -341,6 +354,53 @@ void Kleo::CryptoConfigModule::cancel()
 
 ////
 
+namespace
+{
+bool offerEntryForConfiguration(QGpgME::CryptoConfigEntry *entry)
+{
+    static const QRegularExpression entryPathGroupSegmentRegexp{QStringLiteral("/.*/")};
+
+    static std::set<QString> entriesToExclude;
+    if (entriesToExclude.empty()) {
+        entriesToExclude.insert(QStringLiteral("gpg/keyserver"));
+        if (engineIsVersion(2, 3, 5, GpgME::GpgConfEngine)
+            || (engineIsVersion(2, 2, 34, GpgME::GpgConfEngine)
+                && !engineIsVersion(2, 3, 0, GpgME::GpgConfEngine))) {
+            // exclude for 2.2.{34,...} and 2.3.5+
+            entriesToExclude.insert(QStringLiteral("gpgsm/keyserver"));
+        }
+    }
+
+    const bool de_vs = Kleo::gnupgUsesDeVsCompliance();
+    // Skip "dangerous" expert options if we are running in CO_DE_VS.
+    // Otherwise, skip any options beyond "invisible" (== expert + 1) level.
+    const auto maxEntryLevel = de_vs ? QGpgME::CryptoConfigEntry::Level_Advanced
+                                     : QGpgME::CryptoConfigEntry::Level_Expert + 1;
+    // we ignore the group when looking up entries to exclude because entries
+    // are uniquely identified by their name and their component
+    const auto entryId = entry->path().replace(entryPathGroupSegmentRegexp, QLatin1String{"/"}).toLower();
+    return (entry->level() <= maxEntryLevel) &&
+           (entriesToExclude.find(entryId) == entriesToExclude.end());
+}
+
+auto getGroupEntriesToOfferForConfiguration(QGpgME::CryptoConfigGroup *group)
+{
+
+    std::vector<QGpgME::CryptoConfigEntry *> result;
+    const auto entryNames = group->entryList();
+    for (const auto &entryName : entryNames) {
+        auto *const entry = group->entry(entryName);
+        Q_ASSERT(entry);
+        if (offerEntryForConfiguration(entry)) {
+            result.push_back(entry);
+        } else {
+            qCDebug(KLEO_UI_LOG) << "entry" << entry->path() << "too advanced or excluded explicitly, skipping";
+        }
+    }
+    return result;
+}
+}
+
 Kleo::CryptoConfigComponentGUI::CryptoConfigComponentGUI(
     CryptoConfigModule *module, QGpgME::CryptoConfigComponent *component,
     QWidget *parent)
@@ -357,14 +417,23 @@ Kleo::CryptoConfigComponentGUI::CryptoConfigComponentGUI(
             if (!group) {
                 continue;
             }
+            auto groupEntries = getGroupEntriesToOfferForConfiguration(group);
+            if (groupEntries.size() == 0) {
+                // skip groups without entries to be offered in the UI
+                continue;
+            }
             const QString title = group->description();
             auto hl = new KDHorizontalLine(title.isEmpty() ? *it : title, this);
             const int row = glay->rowCount();
             glay->addWidget(hl, row, 0, 1, 3);
-            mGroupGUIs.append(new CryptoConfigGroupGUI(module, group, glay, this));
+            mGroupGUIs.append(new CryptoConfigGroupGUI(module, group, groupEntries, glay, this));
         }
     } else if (!groups.empty()) {
-        mGroupGUIs.append(new CryptoConfigGroupGUI(module, mComponent->group(groups.front()), glay, this));
+        auto *const group = mComponent->group(groups.front());
+        auto groupEntries = getGroupEntriesToOfferForConfiguration(group);
+        if (groupEntries.size() > 0) {
+            mGroupGUIs.append(new CryptoConfigGroupGUI(module, group, groupEntries, glay, this));
+        }
     }
     glay->setRowStretch(glay->rowCount(), 1);
 }
@@ -400,23 +469,17 @@ void Kleo::CryptoConfigComponentGUI::defaults()
 ////
 
 Kleo::CryptoConfigGroupGUI::CryptoConfigGroupGUI(
-    CryptoConfigModule *module, QGpgME::CryptoConfigGroup *group,
-    QGridLayout *glay, QWidget *widget)
-    : QObject(module), mGroup(group)
+        CryptoConfigModule *module,
+        QGpgME::CryptoConfigGroup *group,
+        const std::vector<QGpgME::CryptoConfigEntry *> &entries,
+        QGridLayout *glay,
+        QWidget *widget)
+    : QObject(module)
 {
-    const bool de_vs = Kleo::Formatting::complianceMode() == QLatin1String("de-vs");
     const int startRow = glay->rowCount();
-    const QStringList entries = mGroup->entryList();
-    for (QStringList::const_iterator it = entries.begin(), end = entries.end(); it != end; ++it) {
-        QGpgME::CryptoConfigEntry *entry = group->entry(*it);
-        Q_ASSERT(entry);
-        /* Skip "dangerous" options if we are running in CO_DE_VS.  */
-        if (de_vs && entry->level() > QGpgME::CryptoConfigEntry::Level_Advanced) {
-            qCDebug(KLEO_UI_LOG) << "entry" << *it << "too advanced, skipping";
-            continue;
-        }
+    for (auto entry : entries) {
         CryptoConfigEntryGUI *entryGUI =
-            CryptoConfigEntryGUIFactory::createEntryGUI(module, entry, *it, glay, widget);
+            CryptoConfigEntryGUIFactory::createEntryGUI(module, entry, entry->name(), glay, widget);
         if (entryGUI) {
             mEntryGUIs.append(entryGUI);
             entryGUI->load();
@@ -484,7 +547,7 @@ static const struct WidgetsByEntryName {
     constructor create;
 } widgetsByEntryName[] = {
     { "*/*/debug-level",   &_create<CryptoConfigEntryDebugLevel> },
-    { "gpg/*/keyserver",   &_create<CryptoConfigEntryKeyserver>  }
+    { "scdaemon/*/reader-port", &_create<CryptoConfigEntryReaderPort> },
 };
 static const unsigned int numWidgetsByEntryName = sizeof widgetsByEntryName / sizeof * widgetsByEntryName;
 
@@ -518,21 +581,22 @@ CryptoConfigEntryGUI *Kleo::CryptoConfigEntryGUIFactory::createEntryGUI(CryptoCo
 
     // try to lookup by path:
     const QString path = entry->path();
-    for (unsigned int i = 0; i < numWidgetsByEntryName; ++i)
+    for (unsigned int i = 0; i < numWidgetsByEntryName; ++i) {
         if (QRegExp(QLatin1String(widgetsByEntryName[i].entryGlob), Qt::CaseSensitive, QRegExp::Wildcard).exactMatch(path)) {
             return widgetsByEntryName[i].create(module, entry, entryName, glay, widget);
         }
+    }
 
     // none found, so look up by type:
     const unsigned int argType = entry->argType();
     Q_ASSERT(argType < QGpgME::CryptoConfigEntry::NumArgType);
-    if (entry->isList())
+    if (entry->isList()) {
         if (const constructor create = listWidgets[argType]) {
             return create(module, entry, entryName, glay, widget);
         } else {
             qCWarning(KLEO_UI_LOG) << "No widget implemented for list of type" << entry->argType();
         }
-    else if (const constructor create = scalarWidgets[argType]) {
+    } else if (const constructor create = scalarWidgets[argType]) {
         return create(module, entry, entryName, glay, widget);
     } else {
         qCWarning(KLEO_UI_LOG) << "No widget implemented for type" << entry->argType();
@@ -611,16 +675,18 @@ void Kleo::CryptoConfigEntryLineEdit::doLoad()
 }
 
 ////
-
+/* Note: Do not use "guru" as debug level but use the value 10.  The
+   former also enables the creation of hash dump files and thus leaves
+   traces of plaintext on the disk.  */
 static const struct {
-    const char *label;
+    const KLazyLocalizedString label;
     const char *name;
 } debugLevels[] = {
-    { I18N_NOOP("0 - None"), "none"},
-    { I18N_NOOP("1 - Basic"), "basic"},
-    { I18N_NOOP("2 - Verbose"), "advanced"},
-    { I18N_NOOP("3 - More Verbose"), "expert"},
-    { I18N_NOOP("4 - All"), "guru"},
+    { kli18n("0 - None"), "none"},
+    { kli18n("1 - Basic"), "basic"},
+    { kli18n("2 - Verbose"), "advanced"},
+    { kli18n("3 - More Verbose"), "expert"},
+    { kli18n("4 - All"), "10"},
 };
 static const unsigned int numDebugLevels = sizeof debugLevels / sizeof * debugLevels;
 
@@ -633,14 +699,14 @@ Kleo::CryptoConfigEntryDebugLevel::CryptoConfigEntryDebugLevel(CryptoConfigModul
     label->setBuddy(mComboBox);
 
     for (unsigned int i = 0; i < numDebugLevels; ++i) {
-        mComboBox->addItem(i18n(debugLevels[i].label));
+        mComboBox->addItem(KLocalizedString(debugLevels[i].label).toString());
     }
 
     if (entry->isReadOnly()) {
         label->setEnabled(false);
         mComboBox->setEnabled(false);
     } else {
-        connect(mComboBox, QOverload<int>::of(&QComboBox::currentIndexChanged),
+        connect(mComboBox, qOverload<int>(&QComboBox::currentIndexChanged),
                 this, &CryptoConfigEntryDebugLevel::slotChanged);
     }
 
@@ -662,11 +728,12 @@ void Kleo::CryptoConfigEntryDebugLevel::doSave()
 void Kleo::CryptoConfigEntryDebugLevel::doLoad()
 {
     const QString str = mEntry->stringValue();
-    for (unsigned int i = 0; i < numDebugLevels; ++i)
+    for (unsigned int i = 0; i < numDebugLevels; ++i) {
         if (str == QLatin1String(debugLevels[i].name)) {
             mComboBox->setCurrentIndex(i);
             return;
         }
+    }
     mComboBox->setCurrentIndex(0);
 }
 
@@ -775,7 +842,7 @@ Kleo::CryptoConfigEntrySpinBox::CryptoConfigEntrySpinBox(
     } else {
         mNumInput->setMinimum(mKind == Int ? std::numeric_limits<int>::min() : 0);
         mNumInput->setMaximum(std::numeric_limits<int>::max());
-        connect(mNumInput, QOverload<int>::of(&QSpinBox::valueChanged), this, &CryptoConfigEntrySpinBox::slotChanged);
+        connect(mNumInput, qOverload<int>(&QSpinBox::valueChanged), this, &CryptoConfigEntrySpinBox::slotChanged);
     }
 }
 
@@ -900,22 +967,34 @@ void prepareURLCfgDialog(QDialog *dialog, DirectoryServicesWidget *dirserv, bool
 
 void Kleo::CryptoConfigEntryLDAPURL::slotOpenDialog()
 {
+    if (!gpgme_check_version("1.16.0")) {
+        KMessageBox::sorry(mPushButton->parentWidget(),
+                           i18n("Configuration of directory services is not possible "
+                                "because the used gpgme libraries are too old."),
+                           i18n("Sorry"));
+        return;
+    }
+
     // I'm a bad boy and I do it all on the stack. Enough classes already :)
     // This is just a simple dialog around the directory-services-widget
     QDialog dialog(mPushButton->parentWidget());
-    dialog.setWindowTitle(i18nc("@title:window", "Configure LDAP Servers"));
+    dialog.setWindowTitle(i18nc("@title:window", "Configure Directory Services"));
 
     auto dirserv = new DirectoryServicesWidget(&dialog);
 
     prepareURLCfgDialog(&dialog, dirserv, mEntry->isReadOnly());
 
-    dirserv->setX509ReadOnly(mEntry->isReadOnly());
-    dirserv->setAllowedSchemes(DirectoryServicesWidget::LDAP);
-    dirserv->setAllowedProtocols(DirectoryServicesWidget::X509Protocol);
-    dirserv->addX509Services(mURLList);
+    dirserv->setReadOnly(mEntry->isReadOnly());
+
+    std::vector<KeyserverConfig> servers;
+    std::transform(std::cbegin(mURLList), std::cend(mURLList), std::back_inserter(servers), [](const auto &url) { return KeyserverConfig::fromUrl(url); });
+    dirserv->setKeyservers(servers);
 
     if (dialog.exec()) {
-        setURLList(dirserv->x509Services());
+        QList<QUrl> urls;
+        const auto servers = dirserv->keyservers();
+        std::transform(std::begin(servers), std::end(servers), std::back_inserter(urls), [](const auto &server) { return server.toUrl(); });
+        setURLList(urls);
         slotChanged();
     }
 }
@@ -930,122 +1009,4 @@ void Kleo::CryptoConfigEntryLDAPURL::setURLList(const QList<QUrl> &urlList)
     }
 }
 
-Kleo::CryptoConfigEntryKeyserver::CryptoConfigEntryKeyserver(
-    CryptoConfigModule *module,
-    QGpgME::CryptoConfigEntry *entry,
-    const QString &entryName,
-    QGridLayout *glay, QWidget *widget)
-    : CryptoConfigEntryGUI(module, entry, entryName)
-{
-    mLabel = new QLabel(widget);
-    mPushButton = new QPushButton(i18n("Edit..."), widget);
-
-    const int row = glay->rowCount();
-    QLabel *label = new QLabel(i18n("Use keyserver at"), widget);
-    label->setBuddy(mPushButton);
-    glay->addWidget(label, row, 1);
-    auto hlay = new QHBoxLayout;
-    glay->addLayout(hlay, row, 2);
-    hlay->addWidget(mLabel, 1);
-    hlay->addWidget(mPushButton);
-
-    if (entry->isReadOnly()) {
-        mLabel->setEnabled(false);
-        mPushButton->hide();
-    } else {
-        connect(mPushButton, &QPushButton::clicked, this, &CryptoConfigEntryKeyserver::slotOpenDialog);
-    }
-}
-
-Kleo::ParsedKeyserver Kleo::parseKeyserver(const QString &str)
-{
-    const QStringList list = str.split(QRegExp(QLatin1String("[\\s,]")), Qt::SkipEmptyParts);
-    if (list.empty()) {
-        return Kleo::ParsedKeyserver();
-    }
-    Kleo::ParsedKeyserver result;
-    result.url = list.front();
-    const auto listMid = list.mid(1);
-    for (const QString &kvpair : listMid) {
-        const int idx = kvpair.indexOf(QLatin1Char('='));
-        if (idx < 0) {
-            result.options.push_back(qMakePair(kvpair, QString()));     // null QString
-        } else {
-            const QString key = kvpair.left(idx);
-            const QString value = kvpair.mid(idx + 1);
-            if (value.isEmpty()) {
-                result.options.push_back(qMakePair(key, QStringLiteral("")));    // make sure it's not a null QString, only an empty one
-            } else {
-                result.options.push_back(qMakePair(key, value));
-            }
-        }
-    }
-    return result;
-}
-
-QString Kleo::assembleKeyserver(const ParsedKeyserver &keyserver)
-{
-    if (keyserver.options.empty()) {
-        return keyserver.url;
-    }
-    QString result = keyserver.url;
-    typedef QPair<QString, QString> Pair;
-    for (const Pair &pair : qAsConst(keyserver.options))
-        if (pair.second.isNull()) {
-            result += QLatin1Char(' ') + pair.first;
-        } else {
-            result += QLatin1Char(' ') + pair.first + QLatin1Char('=') + pair.second;
-        }
-    return result;
-}
-
-void Kleo::CryptoConfigEntryKeyserver::doLoad()
-{
-    mParsedKeyserver = parseKeyserver(mEntry->stringValue());
-    mLabel->setText(mParsedKeyserver.url);
-}
-
-void Kleo::CryptoConfigEntryKeyserver::doSave()
-{
-    mParsedKeyserver.url = mLabel->text();
-    mEntry->setStringValue(assembleKeyserver(mParsedKeyserver));
-}
-
-static QList<QUrl> string2urls(const QString &str)
-{
-    QList<QUrl> ret;
-    if (str.isEmpty()) {
-        return ret;
-    }
-    ret << QUrl::fromUserInput(str);
-    return ret;
-}
-
-static QString urls2string(const QList<QUrl> &urls)
-{
-    return urls.empty() ? QString() : urls.front().url();
-}
-
-void Kleo::CryptoConfigEntryKeyserver::slotOpenDialog()
-{
-    // I'm a bad boy and I do it all on the stack. Enough classes already :)
-    // This is just a simple dialog around the directory-services-widget
-    QDialog dialog(mPushButton->parentWidget());
-    dialog.setWindowTitle(i18nc("@title:window", "Configure Keyservers"));
-
-    auto dirserv = new DirectoryServicesWidget(&dialog);
-
-    prepareURLCfgDialog(&dialog, dirserv, mEntry->isReadOnly());
-
-    dirserv->setOpenPGPReadOnly(mEntry->isReadOnly());
-    dirserv->setAllowedSchemes(DirectoryServicesWidget::AllSchemes);
-    dirserv->setAllowedProtocols(DirectoryServicesWidget::OpenPGPProtocol);
-    dirserv->addOpenPGPServices(string2urls(mLabel->text()));
-
-    if (dialog.exec()) {
-        mLabel->setText(urls2string(dirserv->openPGPServices()));
-        slotChanged();
-    }
-}
-
 #include "moc_cryptoconfigmodule_p.cpp"
